package me.mrletsplay.streamdeck.action;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import me.mrletsplay.mrcore.json.JSONObject;
import me.mrletsplay.mrcore.json.JSONType;
import me.mrletsplay.simplehttpserver.http.validation.JsonObjectValidator;
import me.mrletsplay.streamdeck.action.impl.PlaySoundAction;
import me.mrletsplay.streamdeck.action.impl.PressKeyAction;
import me.mrletsplay.streamdeck.exception.InitializationException;

public class ActionSerializer {

	public static final JsonObjectValidator
		ACTION_VALIDATOR = new JsonObjectValidator()
			.require("type", JSONType.STRING)
			.require("data", JSONType.OBJECT);

	private static final List<Class<? extends DeckAction>> ACTIONS = List.of(
		PlaySoundAction.class,
		PressKeyAction.class
	);

	private static Map<String, Class<? extends DeckAction>> nameToAction;

	public static void initialize() throws InitializationException {
		Map<String, Class<? extends DeckAction>> nta = new LinkedHashMap<>();
		for(var actionClass : ACTIONS) {
			Action ac = actionClass.getAnnotation(Action.class);
			if(ac == null) throw new InitializationException("Class " + actionClass.getName() + " is missing the @Action annotation");

			nta.put(ac.value(), actionClass);
		}

		nameToAction = Collections.unmodifiableMap(nta);
	}

	public static JSONObject serializeAction(DeckAction action) throws ActionSerializationException {
		if(nameToAction == null) throw new ActionSerializationException("ActionSerializer is not initialized");

		Class<? extends DeckAction> actionClass = action.getClass();
		if(!ACTIONS.contains(actionClass)) throw new ActionSerializationException("Class " + actionClass.getName() + " is not a registered action class");

		Action ac = actionClass.getAnnotation(Action.class);

		JSONObject object = new JSONObject();
		object.put("type", ac.value());

		JSONObject data = new JSONObject();
		for(Field f : action.getClass().getDeclaredFields()) {
			ActionParameter parameter = f.getAnnotation(ActionParameter.class);
			if(parameter == null) continue;

			String name = !parameter.name().isEmpty() ? parameter.name() : f.getName();
			try {
				f.setAccessible(true);
				data.put(name, f.get(action));
			} catch (IllegalArgumentException | IllegalAccessException e) {
				throw new ActionSerializationException(e);
			}
		}

		action.serialize(data);

		return object;
	}

	public static DeckAction deserializeAction(JSONObject action) throws ActionSerializationException {
		if(nameToAction == null) throw new ActionSerializationException("ActionSerializer is not initialized");

		String actionType = action.optString("type").orElseThrow(() -> new ActionSerializationException("Missing action type"));
		Class<? extends DeckAction> actionClass = nameToAction.get(actionType);
		if(actionClass == null) throw new ActionSerializationException("Action doesn't exist");

		JSONObject data = action.optJSONObject("data").orElseThrow(() -> new ActionSerializationException("Missing data"));

		try {
			Constructor<? extends DeckAction> constr = actionClass.getDeclaredConstructor();
			constr.setAccessible(true);
			DeckAction a = constr.newInstance();

			for(Field f : action.getClass().getDeclaredFields()) {
				ActionParameter parameter = f.getAnnotation(ActionParameter.class);
				if(parameter == null) continue;

				String name = !parameter.name().isEmpty() ? parameter.name() : f.getName();
				try {
					f.setAccessible(true);

					// TODO: allow JSONSerializable
					f.set(a, JSONType.castJSONValueTo(data.get(name), f.getType()));
				} catch (IllegalArgumentException | IllegalAccessException e) {
					throw new ActionSerializationException(e);
				}
			}

			a.deserialize(data);
		} catch (InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException | NoSuchMethodException | SecurityException e) {
			throw new ActionSerializationException(e);
		}


		return null;
	}

}
