package me.mrletsplay.streamdeck.api;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Base64;

import javax.imageio.ImageIO;

import me.mrletsplay.mrcore.json.JSONArray;
import me.mrletsplay.mrcore.json.JSONObject;
import me.mrletsplay.mrcore.json.JSONType;
import me.mrletsplay.simplehttpserver.http.HttpRequestMethod;
import me.mrletsplay.simplehttpserver.http.HttpStatusCodes;
import me.mrletsplay.simplehttpserver.http.endpoint.Endpoint;
import me.mrletsplay.simplehttpserver.http.endpoint.EndpointCollection;
import me.mrletsplay.simplehttpserver.http.header.DefaultClientContentTypes;
import me.mrletsplay.simplehttpserver.http.request.HttpRequestContext;
import me.mrletsplay.simplehttpserver.http.response.JsonResponse;
import me.mrletsplay.simplehttpserver.http.validation.JsonArrayValidator;
import me.mrletsplay.simplehttpserver.http.validation.JsonObjectValidator;
import me.mrletsplay.simplehttpserver.http.validation.result.ValidationResult;
import me.mrletsplay.streamdeck.StreamDeck;
import me.mrletsplay.streamdeck.action.ActionSerializationException;
import me.mrletsplay.streamdeck.action.ActionSerializer;
import me.mrletsplay.streamdeck.action.DeckAction;
import me.mrletsplay.streamdeck.deck.DeckButton;

public class StreamDeckAPI implements EndpointCollection {

	private static final JsonObjectValidator
		BUTTON_VALIDATOR = new JsonObjectValidator()
			.require("id", JSONType.INTEGER)
			.optional("icon", JSONType.STRING)
			.optionalObject("action", ActionSerializer.ACTION_VALIDATOR),
		STATE_VALIDATOR = new JsonObjectValidator()
			.requireArray("buttons", new JsonArrayValidator()
				.requireElementObjects(BUTTON_VALIDATOR));

	private static JsonResponse error(String errorMessage) {
		JSONObject error = new JSONObject();
		error.put("error", errorMessage);
		return new JsonResponse(error);
	}

	@Endpoint(method = HttpRequestMethod.GET, path = "/actions")
	public void getActions() {
		// TODO
	}

	@Endpoint(method = HttpRequestMethod.GET, path = "/state")
	public void getState() {
		HttpRequestContext ctx = HttpRequestContext.getCurrentContext();
		JSONObject state = new JSONObject();

		JSONArray buttons = new JSONArray();
		for(DeckButton button : StreamDeck.getDeck().getButtons()) {
			JSONObject buttonObj = new JSONObject();
			buttonObj.put("id", button.getId());

			if(button.getIcon() == null) {
				buttonObj.put("icon", null);
			}else {
				try {
					ByteArrayOutputStream bOut = new ByteArrayOutputStream();
					ImageIO.write(button.getIcon(), "PNG", bOut);
				}catch(IOException e) {
					ctx.respond(HttpStatusCodes.INTERNAL_SERVER_ERROR_500, error("Failed to serialize icon: " + e.getMessage()));
					return;
				}
			}

			if(button.getAction() == null) {
				buttonObj.put("action", null);
			}else {
				try {
					buttonObj.put("action", ActionSerializer.serializeAction(button.getAction()));
				}catch(ActionSerializationException e) {
					ctx.respond(HttpStatusCodes.INTERNAL_SERVER_ERROR_500, error("Failed to serialize action " + e.getMessage()));
					return;
				}
			}

			buttons.add(buttonObj);
		}

		state.put("buttons", buttons);

		ctx.respond(HttpStatusCodes.OK_200, new JsonResponse(buttons));
	}

	@Endpoint(method = HttpRequestMethod.PATCH, path = "/state")
	public void updateState() {
		HttpRequestContext ctx = HttpRequestContext.getCurrentContext();
		JSONObject content;
		if((content = ctx.expectContent(DefaultClientContentTypes.JSON_OBJECT)) == null) {
			ctx.respond(HttpStatusCodes.BAD_REQUEST_400, error("Invalid content"));
			return;
		}

		ValidationResult result = STATE_VALIDATOR.validate(content);
		if(!result.isOk()) {
			ctx.respond(HttpStatusCodes.BAD_REQUEST_400, result.asJsonResponse());
			return;
		}

		for(Object o : content.getJSONArray("buttons")) {
			JSONObject buttonObj = (JSONObject) o;
			int id = buttonObj.getInt("id");
			DeckButton button = StreamDeck.getDeck().getButton(id);

			if(buttonObj.has("icon")) {
				String icon = buttonObj.getString("icon");
				if(icon == null) {
					button.setIcon(null);
				}else {
					try {
						button.setIcon(ImageIO.read(new ByteArrayInputStream(Base64.getDecoder().decode(icon))));
					} catch (IOException | IllegalArgumentException e) {
						ctx.respond(HttpStatusCodes.BAD_REQUEST_400, error("Invalid icon"));
					}
				}
			}

			if(buttonObj.has("action")) {
				JSONObject action = buttonObj.getJSONObject("action");
				if(action == null) {
					button.setAction(null);
				}else {
					DeckAction newAction;
					try {
						newAction = ActionSerializer.deserializeAction(action);
					} catch (ActionSerializationException e) {
						ctx.respond(HttpStatusCodes.BAD_REQUEST_400, error("Invalid action: " + e.getMessage()));
						return;
					}

					button.setAction(newAction);
				}
			}
		}
	}

	@Override
	public String getBasePath() {
		return "/api";
	}

}
